//==============================================================================
// CLASS DECLARATION
//==============================================================================
/** @class
    @private */
net.user1.orbiter.CoreMessageListener = function (orbiter) {
  this.orbiter = orbiter;
  this.log = orbiter.getLog();      
  this.registerCoreListeners();
  this.orbiter.getConnectionManager().addEventListener(net.user1.orbiter.ConnectionManagerEvent.SELECT_CONNECTION, 
                                                       this.selectConnectionListener, this);
};

net.user1.orbiter.CoreMessageListener.prototype.registerCoreListeners = function () {
  var msgMan = this.orbiter.getMessageManager();
  msgMan.addMessageListener(net.user1.orbiter.UPC.RECEIVE_MESSAGE, this.u7, this);
  msgMan.addMessageListener(net.user1.orbiter.UPC.SERVER_HELLO, this.u66, this);
  msgMan.addMessageListener(net.user1.orbiter.UPC.SESSION_TERMINATED, this.u84, this);
};

net.user1.orbiter.CoreMessageListener.prototype.selectConnectionListener = function (e) {
  var msgMan = this.orbiter.getMessageManager();
  if (msgMan.removeListenersOnDisconnect) {
    this.registerCoreListeners();
  }
};

net.user1.orbiter.CoreMessageListener.prototype.u7 = function (message,
                                                               broadcastType,
                                                               fromClientID,
                                                               toRoomID) {
  var msgMan = this.orbiter.getMessageManager();
  var args;
  var userDefinedArgs = Array.prototype.slice.call(arguments).slice(4);

  // ===== To Clients, or To Server =====
  if (broadcastType != net.user1.orbiter.ReceiveMessageBroadcastType.TO_ROOMS) {
    args = [fromClientID].concat(userDefinedArgs);
    msgMan.notifyMessageListeners(message, args);
    return;
  }
  
  // ===== To Rooms =====
  var listeners = msgMan.getMessageListeners(message);

  // Split the recipient room ID into two parts
  var toRoomSimpleID  = net.user1.orbiter.RoomIDParser.getSimpleRoomID(toRoomID);
  var toRoomQualifier = net.user1.orbiter.RoomIDParser.getQualifier(toRoomID);
  var listenerFound; 
  var listenerIgnoredMessage;
  var messageListener;
                                       
  // ===== Run once for each message listener =====
  for (var i = 0; i < listeners.length; i++) {
    messageListener = listeners[i];
    listenerIgnoredMessage = true;
    
    // --- Has no "forRoomIDs" filter ---
    if (messageListener.getForRoomIDs() == null) {
      args = [fromClientID, toRoomID].concat(userDefinedArgs);
      messageListener.getListenerFunction().apply(messageListener.getThisArg(), args);
      listenerFound = true;
      listenerIgnoredMessage = false;
      continue;  // Done with this listener. On to the next.
    }
    
    // --- Has a "forRoomIDs" filter ---
    var listenerRoomIDs = messageListener.getForRoomIDs();
    var listenerRoomQualifier;
    var listenerRoomSimpleID;
    var listenerRoomIDString;
    // ===== Run once for each room id =====
    for (var i = 0; i < listenerRoomIDs.length; i++) {
      listenerRoomIDString = listenerRoomIDs[i];
      listenerRoomQualifier = net.user1.orbiter.RoomIDParser.getQualifier(listenerRoomIDString);
      listenerRoomSimpleID  = net.user1.orbiter.RoomIDParser.getSimpleRoomID(listenerRoomIDString);

      // Check if the listener is interested in the recipient room...
      if (listenerRoomQualifier == toRoomQualifier
          && 
          (listenerRoomSimpleID == toRoomSimpleID
           || listenerRoomSimpleID == "*")) {
        // Found a match. Notify the listener...
          
        // Prepare args.
        if (listenerRoomIDs.length == 1) {
          // The listener is interested in messages sent to a 
          // specific room only, so omit the "toRoom" arg.
          args = [fromClientID].concat(userDefinedArgs);
        } else {
          // The listener is interested in messages sent to 
          // multiple rooms, so include the "toRoomID" arg so the listener 
          // knows which room received the message.
          args = [fromClientID, toRoomID].concat(userDefinedArgs);
        }
        
        messageListener.getListenerFunction().apply(messageListener.getThisArg(), args);
        listenerFound = true;
        listenerIgnoredMessage = false;
        break; // Stop looking at this listener's room ids
      }
    } // Done looking at this listener's room ids
    if (listenerIgnoredMessage) {
      this.log.debug("Message listener ignored message: " + message + ". "
                     + "Listener registered to receive " 
                     + "messages sent to: " + messageListener.getForRoomIDs() 
                     + ", but message was sent to: " + toRoomID);
    }
  } // Done looking at listeners for the incoming message
  if (!listenerFound) {
    this.log.warn("No message listener handled incoming message: " 
                  + message + ", sent to: " + toRoomID);
  }
};

net.user1.orbiter.CoreMessageListener.prototype.u66 = function (serverVersion, 
                                                                sessionID,
                                                                serverUPCVersionString,
                                                                protocolCompatible,
                                                                affinityAddress,
                                                                affinityDuration) {
  this.log.info("[ORBITER] Server version: " + serverVersion);
  this.log.info("[ORBITER] Server UPC version: " + serverUPCVersionString);
  
  var inProgressConnection = this.orbiter.getConnectionManager().getInProgressConnection();
  var inProgressConnectionHost = inProgressConnection.getHost();
  if (affinityAddress != ""
      && typeof affinityAddress !== "undefined"  
      && affinityAddress != inProgressConnectionHost) {
    this.orbiter.getConnectionManager().setAffinity(inProgressConnectionHost, 
                                                    affinityAddress, 
                                                    parseFloat(affinityDuration));
    inProgressConnection.applyAffinity();
  }
};

/**
 * SESSION_TERMINATED
 */
net.user1.orbiter.CoreMessageListener.prototype.u84 = function () {
  this.orbiter.getConnectionManager().dispatchSessionTerminated();
};
