//==============================================================================
// CLASS DECLARATION
//==============================================================================
/** @class
 * The Orbiter class is the root class of every OrbiterMicro application.
 * It provides basic tools for connecting to Union server, and gives
 * the application access to the core Orbiter system modules.
 * The Orbiter class dispatches the following events:

<ul class="summary">
<li class="fixedFont">{@link net.user1.orbiter.OrbiterEvent.READY}</li>
<li class="fixedFont">{@link net.user1.orbiter.OrbiterEvent.CLOSE}</li>
<li class="fixedFont">{@link net.user1.orbiter.OrbiterEvent.PROTOCOL_INCOMPATIBLE}</li>
</ul>

To register for events, use {@link net.user1.events.EventDispatcher#addEventListener}.

    @extends net.user1.events.EventDispatcher
*/
net.user1.orbiter.Orbiter = function () {
  // Invoke superclass constructor
  net.user1.events.EventDispatcher.call(this);
  
  // Initialization. For non-browser environments, set window to null.
  this.window = typeof window == "undefined" ? null : window;
  
  // Initialize system versions.
  this.system = new net.user1.orbiter.System(this.window);
  this.log = new net.user1.logger.Logger();
  
  // Output host version information.
  if (typeof navigator != "undefined") {
    this.log.info("User Agent: " + navigator.userAgent + " " + navigator.platform);
  }
  this.log.info("Union Client Version: " + this.system.getClientType() + " " + this.system.getClientVersion().toStringVerbose());
  this.log.info("Client UPC Protocol Version: " + this.system.getUPCVersion().toString());
  this.consoleLogger = null;
  
  if (!this.system.isJavaScriptCompatible()) {
    // Missing required JavaScript capabilities, so abort. 
    this.log.fatal("[ORBITERMICRO] JavaScript version incompatibility detected." 
                   + " Quitting.");
    return;
  }
  
  // Set up the connection manager.
  this.connectionMan = new net.user1.orbiter.ConnectionManager(this);
  
  // Set up the message manager.
  this.messageManager = new net.user1.orbiter.MessageManager(this.log, this.connectionMan);
  
  
  // Set up the core message listener
  this.coreMsgListener = new net.user1.orbiter.CoreMessageListener(this);
  
  // Register for ConnectionManager events.
  this.connectionMan.addEventListener(net.user1.orbiter.ConnectionManagerEvent.READY, 
                                      this._readyListener, this);
  this.connectionMan.addEventListener(net.user1.orbiter.ConnectionManagerEvent.CONNECT_FAILURE, 
                                      this._connectFailureListener, this);
  this.connectionMan.addEventListener(net.user1.orbiter.ConnectionManagerEvent.DISCONNECT, 
                                      this._disconnectListener, this);
  
  // Set up the connection monitor
  this.connectionMonitor = new net.user1.orbiter.ConnectionMonitor(this);
  this.connectionMonitor.restoreDefaults();
  
  this.clientID = "";
  this.sessionID = "";
  
  // Self-client shim
  this.selfClient = {};
  this.selfClient.ping = -1;
  this.selfClient.getPing = function () {
    return this.ping;
  };
  
  // Register to be notified when a new connection is about to be opened
  this.connectionMan.addEventListener(net.user1.orbiter.ConnectionManagerEvent.SELECT_CONNECTION, 
                                      this._selectConnectionListener, this);

  // Enable HTTP failover connections
  this.httpFailoverEnabled = true;
  
  this.log.info("[ORBITERMICRO] Initialization complete.");
};

//==============================================================================
// INHERITANCE
//==============================================================================
net.user1.utils.extend(net.user1.orbiter.Orbiter, net.user1.events.EventDispatcher);

//==============================================================================    
// CLIENT INFO
//==============================================================================    
net.user1.orbiter.Orbiter.prototype.getClientID = function () {
  return this.clientID;
};

net.user1.orbiter.Orbiter.prototype.self = function () {
  return this.selfClient;
}

//==============================================================================    
// SESSION ID
//==============================================================================     
/** @private */
net.user1.orbiter.Orbiter.prototype.setSessionID = function (id) {
  this.sessionID = id;
};

net.user1.orbiter.Orbiter.prototype.getSessionID = function () {
  return this.sessionID == null ? "" : this.sessionID;
};

//==============================================================================    
// UPC LISTENERS
//==============================================================================    
/** @private */
net.user1.orbiter.Orbiter.prototype.u29 = function (clientID) {
  this.clientID = clientID;
};

/** @private */
net.user1.orbiter.Orbiter.prototype.u66 = function (clientID,
                                                    sessionID,
                                                    serverUPCVersionString,
                                                    protocolCompatible,
                                                    affinityAddress,
                                                    affinityDuration) {
  var serverUPCVersion = new net.user1.orbiter.VersionNumber();
  serverUPCVersion.fromVersionString(serverUPCVersionString);
  if (protocolCompatible == "false") {
    this.dispatchProtocolIncompatible(serverUPCVersion);
  }
};

/** @private */
net.user1.orbiter.Orbiter.prototype.u164 = function (reason, description) {
  this.connectionMonitor.setAutoReconnectFrequency(-1);
  this.dispatchConnectRefused(new net.user1.orbiter.ConnectionRefusal(reason, description));
};

//==============================================================================    
// CONNECTION
//==============================================================================
/** @private */   
net.user1.orbiter.Orbiter.prototype.buildConnection = function (host, port, type, sendDelay) {
  var connection;
  
  switch (type) {
    case net.user1.orbiter.ConnectionType.HTTP:
      if (this.system.hasHTTPDirectConnection()) {
        connection = new net.user1.orbiter.HTTPDirectConnection();
      } else {
        connection = new net.user1.orbiter.HTTPIFrameConnection();
      }
      break;

    case net.user1.orbiter.ConnectionType.SECURE_HTTP:
      if (this.system.hasHTTPDirectConnection()) {
        connection = new net.user1.orbiter.SecureHTTPDirectConnection();
      } else {
        connection = new net.user1.orbiter.SecureHTTPIFrameConnection();
      }
      break;

    case net.user1.orbiter.ConnectionType.WEBSOCKET:
      connection = new net.user1.orbiter.WebSocketConnection();
      break;

    case net.user1.orbiter.ConnectionType.SECURE_WEBSOCKET:
      connection = new net.user1.orbiter.SecureWebSocketConnection();
      break;

    default:
      throw new Error("[ORBITER] Error at buildConnection(). Invalid type specified: [" + type + "]");
  }
  
  try {
    connection.setServer(host, port);
  } catch (e) {
    this.log.error("[CONNECTION] " + connection.toString() + " " + e);
  } finally {
    this.connectionMan.addConnection(connection);
    if (connection instanceof net.user1.orbiter.HTTPConnection) {
      // Set delay after adding connection so the connection object has
      // access to this Orbiter object
      connection.setSendDelay(sendDelay);
    }
  }
};

/**
 * Connects to Union Server using the specified host and port(s).
 * 
 * @param host
 * @param port1
 * @param port2
 * @param ...
 * @param portn
 */
net.user1.orbiter.Orbiter.prototype.connect = function (host) {
  this.useSecureConnect = false;
  this.doConnect.apply(this, arguments);
};

/**
 * <p>
 * The secureConnect() method is identical to the connect() method, except that
 * it uses an encrypted connection (TLS or SSL) rather than an
 * unencrypted connection. Before secureConnect() can be used, Union Server
 * must be configured to accept client communications over a secure gateway,
 * which includes the installation of a server-side security certificate. For
 * instructions on configuring Union Server for secure communications, see
 * Union Server's documentation at http://unionplatform.com.
 * </p>
 *
 * @see net.user1.orbiter.Orbiter#connect
 */
net.user1.orbiter.Orbiter.prototype.secureConnect = function (host) {
  this.useSecureConnect = true;
  this.doConnect.apply(this, arguments);
};

/**
 * @private
 */
net.user1.orbiter.Orbiter.prototype.doConnect = function (host) {
  var ports = Array.prototype.slice.call(arguments).slice(1);
  if (host != null) {
    this.setServer.apply(this, [host].concat(ports));
  }
  this.log.info("[ORBITER] Connecting to Union...");
  this.connectionMan.connect();
};

net.user1.orbiter.Orbiter.prototype.disconnect = function () {
  this.connectionMan.disconnect();
};

/**
 * Assigns the host and port(s) Orbiter should use when attempting to connect to
 * Union Server. The first argument is the host address (e.g., "example.com"),
 * Subsequent arguments list the integer ports for the connection (e.g., 80). 
 * Orbiter will attempt to connect over the ports in the order specified. For 
 * example, given the code setServer("tryunion.com", 9100, 80, 443, Orbiter
 * will first attempt to connect to Union Server over port 9100; if the
 * connection fails, Orbiter will automatically next attempt to connect over 
 * port 80, if that fails, Orbiter will attempt to connect to port 443. To add 
 * multiple hosts (not just multiple ports) to Orbiter's list of failover
 * connections, use ConnectionManager's addConnection() method.
 * 
 * To reduce network latency and bandwidth consumption, Orbiter automatically
 * attempts to connect via WebSocket wherever WebSocket is supported. Where
 * WebSocket is not supported, Orbiter automatically fails over to HTTP
 * communications.
 * 
 * Wherever possible, to allow maximum connection success by Union clients, 
 * Union Server should be run on port 80.
 * 
 * @param host
 * @param port1
 * @param port2
 * @param ...
 * @param portn
 */
net.user1.orbiter.Orbiter.prototype.setServer = function (host) {
  if (host != null && arguments.length > 1) {
    if (this.connectionMan.getConnections().length > 0) {
      this.connectionMan.removeAllConnections();
    }
    // Where possible, create regular WebSocket connections for the specified
    // host and its ports.
    var connectionType;
    if (this.system.hasWebSocket()) {
      for (var i = 1; i < arguments.length; i++) {
        connectionType = this.useSecureConnect
                         ? net.user1.orbiter.ConnectionType.SECURE_WEBSOCKET
                         : net.user1.orbiter.ConnectionType.WEBSOCKET;
        this.buildConnection(host, arguments[i], connectionType, -1);
      }
    } else {
      this.log.info("[ORBITERMICRO] WebSocket not found in host environment. Trying HTTP.");
    }
    // Next, if failover is enabled or WebSocket is not supported, create HTTPConnections
    if (this.isHTTPFailoverEnabled() || !this.system.hasWebSocket()) {
      for (i = 1; i < arguments.length; i++) {
        connectionType = this.useSecureConnect
                         ? net.user1.orbiter.ConnectionType.SECURE_HTTP
                         : net.user1.orbiter.ConnectionType.HTTP;
        this.buildConnection(host,
                             arguments[i],
                             connectionType,
                             net.user1.orbiter.HTTPConnection.DEFAULT_SEND_DELAY);
      }
    }
  } else {
    this.log.error("[ORBITERMICRO] setServer() failed. Invalid host or port supplied [" + arguments + "].");
  }
};

net.user1.orbiter.Orbiter.prototype.isReady = function () {
  return this.connectionMan.isReady();
};

  
//==============================================================================
// HTTP FAILOVER CONFIGURATION
//==============================================================================
net.user1.orbiter.Orbiter.prototype.enableHTTPFailover = function () {
  this.httpFailoverEnabled = true;
};

net.user1.orbiter.Orbiter.prototype.disableHTTPFailover = function () {
  this.httpFailoverEnabled = false;
};

net.user1.orbiter.Orbiter.prototype.isHTTPFailoverEnabled = function () {
  return this.httpFailoverEnabled;
};

//==============================================================================
// CONNECTION EVENT LISTENERS
//==============================================================================
/** @private */
net.user1.orbiter.Orbiter.prototype._disconnectListener = function (e) {
  this.clientID = "";
  this.dispatchClose();
};

/** @private */
net.user1.orbiter.Orbiter.prototype._connectFailureListener = function (e) {
  this.clientID = "";
  this.dispatchClose();
};

/** @private */
net.user1.orbiter.Orbiter.prototype._readyListener = function (e) {
  this.log.info("[ORBITER] Orbiter now connected and ready.");
  this.dispatchReady();
};

/** @private */
net.user1.orbiter.Orbiter.prototype._selectConnectionListener = function (e) {
  // Register to be notified when the client's ID is received 
  this.messageManager.addMessageListener("u29", this.u29, this);
  // Register to be notified when the server's "hello" message is received 
  this.messageManager.addMessageListener("u66", this.u66, this);
  this.messageManager.addMessageListener(net.user1.orbiter.UPC.CONNECTION_REFUSED, this.u164, this);
};

//==============================================================================    
// INSTANCE RETRIEVAL
//==============================================================================    
net.user1.orbiter.Orbiter.prototype.getLog = function () {
  return this.log;
};

net.user1.orbiter.Orbiter.prototype.getMessageManager = function () {
  return this.messageManager;
};

net.user1.orbiter.Orbiter.prototype.getConnectionManager = function () {
  return this.connectionMan;
};

net.user1.orbiter.Orbiter.prototype.getConnectionMonitor = function () {
  return this.connectionMonitor;
};

net.user1.orbiter.Orbiter.prototype.getSystem = function () {
  return this.system;
};

//==============================================================================    
// LOGGING
//==============================================================================
net.user1.orbiter.Orbiter.prototype.enableConsole = function () {
  if (this.consoleLogger == null) {
    this.consoleLogger = new net.user1.logger.ConsoleLogger(this.log);
  }
}

net.user1.orbiter.Orbiter.prototype.disableConsole = function() {
  if (this.consoleLogger != null) {
    this.consoleLogger.dispose();
    this.consoleLogger = null;
  }
};
  
//==============================================================================    
// EVENT DISPATCH
//==============================================================================
/** @private */
net.user1.orbiter.Orbiter.prototype.dispatchReady = function () {
  this.dispatchEvent(new net.user1.orbiter.OrbiterEvent(net.user1.orbiter.OrbiterEvent.READY));
}

/** @private */
net.user1.orbiter.Orbiter.prototype.dispatchClose = function () {
  this.dispatchEvent(new net.user1.orbiter.OrbiterEvent(net.user1.orbiter.OrbiterEvent.CLOSE));
}

/** @private */
net.user1.orbiter.Orbiter.prototype.dispatchProtocolIncompatible = function () {
  this.dispatchEvent(new net.user1.orbiter.OrbiterEvent(net.user1.orbiter.OrbiterEvent.PROTOCOL_INCOMPATIBLE));
}

/** @private */
net.user1.orbiter.Orbiter.prototype.dispatchConnectRefused = function (refusal) {
  this.dispatchEvent(new net.user1.orbiter.OrbiterEvent(net.user1.orbiter.OrbiterEvent.CONNECT_REFUSED,
                                 null, refusal));
};

//==============================================================================    
// TOSTRING
//==============================================================================
net.user1.orbiter.Orbiter.prototype.toString = function () {
  return "[object Orbiter]";
}
    